"""
monobit.storage.fontformats.fon.os2 - read OS/2 GPI bitmap fonts and LX containers

(c) 2023--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT

this implementation leans heavily on Alexander Taylor's os2font
- http://www.altsan.org/programming/os2font_src.zip
- https://github.com/altsan/os2-gpi-font-tools
"""

from monobit.base import import_all
import_all(__name__)
