"""
monobit.plumbing - command-line and scripting support

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from .scripting import *
from .args import *
# can't import help due to circular dependency
