"""
monobit.storage.fontformats.fon - Windows and OS/2 FON files

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

# ensure formats get registered
from . import fon, windows, os2
