"""
monobit.storage.fontformats.apple - Mac OS and other NFNT-based font formats

(c) 2019--2024 Rob Hagemans
licence: https://opensource.org/licenses/MIT
"""

from monobit.base import import_all
import_all(__name__)
