"""
monobit.storage.fontformats.fon.windows - Windows FON and FNT files

`monobit.storage.fontformats.fon.windows` is copyright 2019--2024 Rob Hagemans
`mkwinfont` is copyright 2001 Simon Tatham. All rights reserved.
`dewinfont` is copyright 2001,2017 Simon Tatham. All rights reserved.

See `LICENSE.md` in this package's directory.
"""

from . import fnt, ne, pe
